"""A sample :class:`zetup.class_package` for testing purposes."""

import zetup

zetup.class_package(__name__, member_modules=['methods'])


class Class(zetup.object):

    def __init__(self):
        pass



import zetup

__all__ = ('class_member_module', )


class class_member_module(zetup.module):

    __package__ = zetup

    @property
    def owner(self):
        return self.__dict__.get('owner')

    @owner.setter
    def owner(self, classobj):
        self.__dict__['owner'] = classobj
        for obj in self.members:
            classobj.member(obj)
        for obj in self.metamembers:
            classobj.metamember(obj)

    def __init__(self, name, api):
        super(class_member_module, self).__init__(
            name, api,
            __iter__=lambda: iter((self.member, self.metamember)))

        self.__dict__['members'] = []
        self.__dict__['metamembers'] = []

    def member(self, obj):
        # assert self.type is not None, (
        #     "{!r} was not correctly registered as member_module of a {!r}"
        #     .format(self, zetup.class_package))
        # return self.type.member(obj)
        self.members.append(obj)

    def metamember(self, obj):
        # assert self.type is not None, (
        #     "{!r} was not correctly registered as member_module of a {!r}"
        #     .format(self, zetup.class_package))
        # return self.type.metamember(obj)
        self.metamembers.append(obj)



"""
Simplify command line argument definitions for ``console_scripts``.

Every according script entry function is termed a :class:`zetup.program` in
this context
"""

from __future__ import absolute_import

import sys
from argparse import ArgumentParser

import zetup

from .object import object

__all__ = ('program', 'with_arguments')


class program(ArgumentParser, object):
    """
    Decorator for entry functions of ``console_scripts``.

    Simplifies argument handling a lot by integration with
    :class:`zetup.with_argument` instead of directly using multiple calls to
    ``argparse.ArgumentParser.add_argument``:

    >>> import zetup
    >>> from zetup import program, with_arguments

    >>> @program(
    ...     with_arguments
    ...     ('--version', action='store_true')
    ...     ('--class')
    ... )
    ... def explain(args):
    ...     if args.version:
    ...         print(zetup.__version__)
    ...         sys.exit(0)
    ...
    ...     from importlib import import_module
    ...     modname, clsname = getattr(args, 'class').rsplit('.', 1)
    ...     mod = import_module(modname)
    ...     print(help(getattr(mod, clsname)))
    ...     sys.exit(0)

    As a (kind of) real life example, the above would of course end the Python
    process. That's why ``SystemExit`` is catched here for demonstration
    purposes, and expicit arg lists are given instead of implicitly using
    ``sys.argv``:

    >>> try:
    ...     explain(['--version'])
    ... except SystemExit as exc:
    ...     assert exc.code == 0
    {zetup_version}

    >>> try:
    ...     explain(['--class', 'zetup.program'])
    ... except SystemExit as exc:
    ...     assert exc.code == 0
    Help on class program in module zetup.program:
    ...
    """

    # the variables in curly braces in the doc string above are substituted
    # below the program class definition

    def __init__(self, arguments, **kwargs):
        """
        Prepare the decorator with a :class:`zetup.with_arguments` object.

        All `kwargs` are delegated to the ``ArgumentParser`` base constructor
        """
        ArgumentParser.__init__(self, **kwargs)
        for args, kwargs in arguments:
            self.add_argument(*args, **kwargs)

    def __call__(self, func):
        """
        Wrap `func` with an automatic :meth:`zetup.program.parse_args` step.

        So that `func` can directly work with the parser result object
        """
        def caller(argv=None):
            """
            Wrap function with auto-parsing of `argv` or default ``sys.argv``.

            Then pass the result to the actual function decorated by
            :class:`zetup.program`
            """
            args = self.parse_args(sys.argv if argv is None else argv)
            return func(args)

        # give wrapper function same name as wrapped one, since it is the one
        # to be returned here
        caller.__name__ = func.__name__
        if hasattr(func, '__qualname__'):
            caller.__qualname__ = func.__qualname__
        return caller

program.__doc__ = program.__doc__.format(
    zetup_version=zetup.__version__)  # pylint: disable=no-member


class with_arguments(object):
    """
    Gather argument definitions for instances of ``argparse.ArgumentParser``.

    To be used with :class:`zetup.program`, which is derived from the latter

    Initialization and further calling support the same signature as
    ``ArgumentParser.add_argument``. Calls always return the instance itself.
    So all definitions can be chained

    The result direclty passed as first positional argument to
    :class:`zetup.program`, which you should also refer to for a complete
    explanation and example

    The ``with_arguments`` name is btw inspired by ``six.with_metaclass`` :)
    """

    def __init__(self, *args, **kwargs):
        """
        Take same arguments as ``argparse.ArgumentParser.add_argument``.

        Initializes the internal list of argument definitions
        """
        self._list = []
        self(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        Take same arguments as ``argparse.ArgumentParser.add_argument``.

        Works like :meth:`.__init__` and extends the internal list of
        argument definitions
        """
        self._list.append((args, kwargs))
        return self

    def __iter__(self):
        return iter(self._list)



# ZETUP | Zimmermann's Extensible Tools for Unified Projects
#
# Copyright (C) 2014-2019 Stefan Zimmermann <user@zimmermann.co>
#
# ZETUP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ZETUP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ZETUP. If not, see <http://www.gnu.org/licenses/>.

"""
Define basic :class:`zetup.object` and its :class:`zetup.meta`.

For extending builtin ``object`` and ``type`` with useful extra features

.. moduleauthor:: Stefan Zimmermann <user@zimmermann.co>
"""

from __future__ import absolute_import

import sys
from inspect import ismodule
from itertools import chain

import zetup

__all__ = ('object', 'meta')


class meta(type):
    """
    Basic metaclass that extends builtin ``type`` with useful extra features.

    Adds a unified basic ``__dir__`` method for Python 2 and 3, which always
    returns all member names from metaclass and class level

    Adds :meth:`.metamethod` and :meth:`.method` decorators for adding new
    members to classes and their metaclasses outside of the class definition
    scopes
    """

    def __init__(cls, clsname, bases, clsattrs):
        super(meta, cls).__init__(clsname, bases, clsattrs)
        if '__package__' not in clsattrs:
            cls.__package__ = None

    if hasattr(type, '__dir__'):  # PY3
        def __dir__(cls):
            """Get all member names from class and metaclass level."""
            return sorted(set(chain(type.__dir__(cls), dir(type(cls)))))

    else:  # PY2
        def __dir__(cls):
            """Get all member names from class and metaclass level."""
            return sorted(set(chain(
                dir(type(cls)), *(c.__dict__ for c in cls.mro()))))

    @classmethod
    def metamember(mcs, obj):
        """
        Decorator for adding `obj` as a member to the metaclass of this class.

        >>> import six  # if you want compatibility with Python 2 and 3
        >>> import zetup

        This is of course only useful when you have a custom metaclass:

        >>> class Meta(zetup.meta):
        ...     pass

        Now you can add new members to your custom metaclass outside of the
        class definition scope:

        >>> @Meta.metamember
        ... def method(cls, arg):
        ...     pass

        >>> Meta.method
        <... Meta.method...>

        This also works from a new class based on this metaclass. In Python 3
        syntax you create such a class like::

            class Class(zetup.object, metaclass=Meta):
                pass

        But if you want to be compatible with Python 2 and 3:

        >>> class Class(six.with_metaclass(Meta, zetup.object)):
        ...     pass

        >>> @Class.metamember
        ... def another_method(cls, arg):
        ...     pass

        >>> Meta.another_method
        <... Meta.another_method...>

        >>> Class.another_method
        <bound method Meta.another_method of ...Class...>

        And the formerly defined ``@Meta.metamember`` is of course also there:

        >>> Class.method
        <bound method Meta.method of ...Class...>
        """
        if isinstance(obj, property):
            name = obj.fget.__name__
        else:
            try:
                name = obj.__name__
            except AttributeError:
                name = obj.__func__.__name__
            else:
                obj.__qualname__ = '%s.%s' % (
                    getattr(mcs, '__qualname__', mcs.__name__),
                    obj.__name__)
        setattr(mcs, name, obj)
        return obj

    def member(cls, obj):
        """
        Decorator for adding `obj` as a member to this class.

        >>> import zetup

        >>> class Class(zetup.object):
        ...     pass

        Now you can add new members to your class outside of the class
        definition scope:

        >>> @Class.member
        ... def method(self, arg):
        ...     pass

        >>> Class.method
        <... Class.method...>

        >>> Class().method
        <bound method Class.method of ...Class...>
        """
        if isinstance(obj, property):
            name = obj.fget.__name__
        else:
            try:
                name = obj.__name__
            except AttributeError:
                name = obj.__func__.__name__
            else:
                obj.__qualname__ = '%s.%s' % (
                    getattr(cls, '__qualname__', cls.__name__),
                    obj.__name__)
        setattr(cls, name, obj)
        return obj

    def __str__(cls):
        clsname = getattr(cls, '__qualname__', cls.__name__)
        mod = getattr(cls, '__package__', None) or cls.__module__
        if ismodule(mod):  # (and not just a string)
            mod = mod.__name__  # pylint: disable=no-member
        elif mod is None:
            return clsname

        return '.'.join((mod, clsname))

    def __repr__(cls):
        """Create PY2/3 unified PY3-style representation."""
        return "<class {!r}>".format(str(cls))


# PY2/3 compatible way to create class `object` with metaclass `meta`...

def __repr__(self):
    """Create PY2/3-unified PY3-style representation."""
    return "<{} at {}>".format(type(self), hex(id(self)).rstrip('L'))

clsattrs = {
    '__package__': zetup,
    '__doc__': """
    Basic class that extends builtin ``object`` with useful extra features.

    Adds a basic ``__dir__`` method for Python 2

    Has :class:`zetup.meta` as metaclass and therefore provides
    :meth:`zetup.meta.metamethod` and :meth:`zetup.meta.method` decorators for
    adding new members to classes and their metaclasses outside of the class
    definition scopes
    """,
    '__repr__': __repr__}

if not hasattr(object, '__dir__'):

    def __dir__(self):
        """Get all member names from instance and class level."""
        return sorted(set(chain(
            self.__dict__,
            *(c.__dict__ for c in type(self).mro()))))

    clsattrs['__dir__'] = __dir__

object = meta('object', (object, ), clsattrs)



"""The parsing framework for ZFG files."""

from __future__ import absolute_import

import re

from six import reraise

from zetup.error import ZetupError
from zetup.object import object, meta

from .data import ZFGData
from .line import BlankLine, OptionLine, SectionLine

__all___ = (
    'ZFG_PARSER', 'ZFGData', 'ZFGParser', 'ZFGSyntaxError', 'parse_zfg')


class ZFGSyntaxError(ZetupError):
    """A ZFG file dosen't follow common INI file syntax."""

    pass


class ZFGParser(object):
    """
    The ZFG file parser.

    ZFG files are actually syntax-compatible with INI files, but their parser
    provides convenient semantic and user-style-friendly extensions, some of
    which are often found in ``.ini`` file handling implementations:

    -   Implicit hierarchies separated by colons in the section names
    -   Sections are partial. Their content can be extended in multiple
        appearances
    -   The resulting parser data stores the exact input file style, including
        original section and option order and even any kind of unnecessary
        white-space, and this is preserved as much as possible when altering
        the data and regenrating text from it

    There is a pre-initialized :const:`zetup.ZFG_PARSER` instance, and its
    :meth:`zetup.ZFGParser.parse` is exposed as :const:`parse_zfg`

    >>> zfgtext = '''
    ... [main]
    ... option = value
    ...
    ... [other]
    ... other option =other value
    ...
    ... [main:sub]
    ... sub option =  sub value
    ... other sub option= other sub value
    ...
    ...
    ... [main]
    ...
    ... further option =   further value
    ... '''.lstrip()

    >>> from zetup import parse_zfg
    >>> zfg = parse_zfg(zfgtext)

    >>> zfg.sections()
    ['main', 'other']

    >>> zfg['main'].options()
    ['option', 'further option']

    >>> zfg['other'].options()
    ['other option']

    >>> zfg['main'].subsections()
    ['sub']

    >>> zfg['main']['sub'].options()
    ['sub option', 'other sub option']
    """

    lineno = None
    match = None
    stack = None

    def parse(self, text, zfg=None):
        if zfg is None:
            zfg = ZFGData()

        def parse_lines():
            if self.lineno > len(text_lines):
                return

            line = text_lines[self.lineno - 1]
            for pattern, handlercls in ZFG_PARSER_RULES:
                match = re.match(pattern, line)
                if match:
                    self.match = match
                    handlercls(parser=self)

                    self.lineno += 1
                    parse_lines()
                    break

            else:
                raise ZFGSyntaxError(self.lineno)

        zfg._lines = []
        self.stack = [zfg]
        self.lineno = 1
        text_lines = text.split('\n')
        parse_lines()
        return zfg


#: The pre-defined :class:`zetup.ZFGParser` instance.
ZFG_PARSER = ZFGParser()


#: The pre-defined function for parsing ZFG files.
#
#  See :class:`zetup.ZFGParser` for more information about ``.zfg`` files
parse_zfg = ZFG_PARSER.parse


#: The regexes and according handlers for parsing ZFG file lines.
#
#  Internally used by :class:`zetup.ZFGParser`
ZFG_PARSER_RULES = [
    (
        r'^\s*$',
        BlankLine),
    (
        r'^\[(?P<name>[^\[\]]+)\]\s*$',
        SectionLine),
    (
        r'^(?P<name>[^\s=][^=]*[^\s=])\s*=\s*(?P<value>.*\S)\s*$',
        OptionLine),
    (
        r'^(?P<name>[^\s=][^=]*[^\s=])\s*=\s*$',
        OptionLine),
]



"""Handlers for parsed ZFG text lines."""

__all__ = ('BlankLine', 'SectionLine', 'OptionLine')


class Line(object):
    """The basic ZFG text line handler."""

    #: The optional identification name of the parsed ZFG file line.
    name = None

    def __init__(self, parser):
        """
        Initialize with :class:`zetup.ZFGParser` instance.

        To get current line number, line text, line regex match object,
        and to access the parser stack
        """
        self.lineno = parser.lineno
        self.text = parser.match.group().rstrip('\r\n')
        parser.stack[-1].lines.append(self)

    def to_zfg(self):
        """Re-create ZFG line text."""
        return self.text

    def __repr__(self):
        """Show line type, line number, and original line text."""
        cls = type(self)
        return "<{} from line {}: {!r}>".format(
            self.name and "{} {!r}".format(cls, self.name) or cls,
            self.lineno, self.text)

Line.__package__ = __package__


class BlankLine(Line):
    """A parsed blank ZFG text line."""

    def __init__(self, parser):
        """Ensure that the line is really blank."""
        super(BlankLine, self).__init__(parser)
        assert not self.text.strip(), (
            "Internal ZFGParser error! Line {} is not really blank: {!r}"
            .format(self.lineno, self.text))


class BlockHeaderLine(Line):
    """
    Base class for parsed ZFG text lines starting data blocks.

    Like sections and options with multi-line values
    """

    #: The parsed sub-lines of a ZFG file's block header line.
    lines = None

    def __init__(self, parser):
        """Initialize the list of the block's sub-lines."""
        super(BlockHeaderLine, self).__init__(parser)
        self.lines = []

    def to_zfg(self):
        """Re-create ZFG data block text."""
        zfg = super(BlockHeaderLine, self).to_zfg()
        if self.lines:
            zfg += "\n{}".format("\n".join(
                line.to_zfg() for line in self.lines))
        return zfg


class SectionLine(BlockHeaderLine):
    """A parsed section header line of ZFG text."""

    def __init__(self, parser):
        """Extract section name and add line handler to the parser stack."""
        if isinstance(parser.stack[-1], OptionLine):
            # remove option header line from parser stack
            optline = parser.stack.pop(-1)
            super_line = parser.stack[-1]
            assert isinstance(super_line, SectionLine), (
                "Interal ZFGParser error! Below {!r} should be a "
                "SectionLine on a the stack, but found {!r}"
                .format(optline, super_line))

        from .data import ZFGData  # avoid circular import
        if isinstance(parser.stack[-1], SectionLine):
            # remove previous section header line from parser stack
            old_line = parser.stack.pop(-1)
            super_line = parser.stack[-1]
            assert isinstance(super_line, ZFGData), (
                "Interal ZFGParser error! Below {!r} should be a ZFGData "
                "object on a the stack, but found {!r}"
                .format(old_line, super_line))

        super(SectionLine, self).__init__(parser)
        self.name = parser.match.group('name')
        parser.stack.append(self)


class OptionLine(BlockHeaderLine):
    """A parsed option header or ``option = value`` line of ZFG text."""

    #: The optional inline value of the ZFG file's option line.
    value = None

    def __init__(self, parser):
        """
        Extract option name and optionally inline option value.

        And add line handler to the parser stack in the non-inline case
        """
        super(OptionLine, self).__init__(parser)
        groups = parser.match.groupdict()
        self.name = groups['name']
        if 'value' in groups:  # ==> inline option value
            self.value = parser.match.group('value')

        else:
            if isinstance(parser.stack[-1], OptionLine):
                # remove previous option header line from parser stack
                old_line = parser.stack.pop(-1)
                super_line = parser.stack[-1]
                assert isinstance(super_line, SectionLine), (
                    "Interal ZFGParser error! Below {!r} should be a "
                    "SectionLine on a the stack, but found {!r}"
                    .format(old_line, super_line))

            parser.stack[-1].lines.append(self)
            parser.stack.append(self)

    def __repr__(self):
        if self.value:
            return "<{} {!r} with value {!r} from line {}: {!r}>".format(
                type(self).__name__, self.name, self.value,
                self.lineno, self.text)

        return super(OptionLine, self).__repr__()



"""All about handling ZFG configuration files."""

from zetup.modules import package

from .data import ZFGData
from .file import ZFGFile
from .parser import ZFG_PARSER, ZFGParser, parse_zfg

package(__name__, (
    'ZFG_PARSER', 'ZFGData', 'ZFGFile', 'ZFGParser', 'parse_zfg'))



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2016 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""zetup.doc

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""

import sys
from importlib import import_module
from inspect import ismodule
from types import ModuleType

import zetup

__all__ = ('AutoDocScopeImporter', )

# zetup.package(__name__, ['AutoDocScopeImporter'])


class EndlessRecursionError(RuntimeError):
    pass


def prevent_endless_recursion(func):
    func.call_stack = []

    def caller(*args, **kwargs):
        if (args, kwargs) in func.call_stack:
            raise EndlessRecursionError(
                "{!r} was called recusively again with args {!r} and "
                "keyword args {!r}".format(func, args, kwargs))

        func.call_stack.append((args, kwargs))
        try:
            result = func(*args, **kwargs)
        finally:
            assert func.call_stack[-1] == (args, kwargs)
            func.call_stack.pop(-1)
        return result

    return caller


class AutoDocScopeModule(ModuleType):

    # to be accepted as packages for subscope imports :)
    __path__ = None

    def __init__(self, name, scope):
        ModuleType.__init__(self, scope.__name__)
        modclass = type(self)

        class Module(type(self)):

            def __getattribute__(self, name):
                if name in ['__class__', '__path__']:
                    return modclass.__getattribute__(self, name)
                return getattr(scope, name)

            def __repr__(self):
                return "<%s for %s>" % (modclass.__name__, repr(scope))

        self.__class__ = Module


class AutoDocScopeImporter(object):

    def __init__(self, pkgname):
        package = import_module(pkgname)
        if not isinstance(package, zetup.package):
            raise TypeError("%s is not a zetup.package instance"
                            % repr(package))
        self.package = package

    def find_module(self, name, path):
        if name.startswith(self.package.__name__ + '.'):
            try:
                if self.load_module(name) is not None:
                    return self

            except EndlessRecursionError:
                pass

    @prevent_endless_recursion
    def load_module(self, name):
        assert name.startswith(self.package.__name__ + '.')
        scope = scopemod = self.package
        for attr in name[len(self.package.__name__) + 1:].split('.'):
            scope = getattr(scope, attr, None)
            if scope is None:
                return None

            if ismodule(scope):
                return scope

            scopemod = AutoDocScopeModule(
                '%s.%s' % (scopemod.__name__, attr), scope)
            scopemod.__path__ = None
            sys.modules[name] = scopemod
        return scope

    # def __repr__(self):
    #     return "<%s for %s>" % (type(self).__name__, repr(self.module))



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

import re
import sys
from textwrap import dedent

import pkg_resources
from pkg_resources import (
    parse_requirements, Requirement, get_distribution)

__all__ = (
    'DistributionNotFound',
    'Requirements',
    'VersionConflict',
    'requires',
)


if sys.version_info[0] == 3:
    unicode = str


def howto_install_requirement(req):
    """Create a message explaining how to install `req`"""
    return dedent("""
    You can install the needed requirement with the following shell command:

        python -m pip install "{}"
    """.format(req)).strip()


class DistributionNotFound(pkg_resources.DistributionNotFound):

    def __init__(self, req, requirer, reason=None):
        super(DistributionNotFound, self).__init__(req, [requirer])
        self.requirer = requirer
        self.reason = reason

    def __str__(self):
        text = "%s needs %s" % (self.requirer, self.req)
        if self.reason:
            text += " (%s)" % self.reason
        return "{}\n\n{}".format(text, howto_install_requirement(self.req))


class VersionConflict(pkg_resources.VersionConflict):

    def __init__(self, req, found_version, requirer, reason=None):
        super(VersionConflict, self).__init__(
            '%s-%s' % (req.key, found_version), req)
        self.requirer = requirer
        self.reason = reason

    def __str__(self):
        text = "%s needs %s but found %s" % (
            self.requirer, self.req, self.dist)
        if self.reason:
            text += " (%s)" % self.reason
        return "{}\n\n{}".format(text, howto_install_requirement(self.req))


class Requirements(object):
    """Package requirements manager."""

    @staticmethod
    def _parse(text):
        """
        Generate parsed requirements from `text`.
        
        :param text:
            Newline-separated requirement specs

        Additionally looks for ``#import name`` comments after requirement
        lines (the actual root module name of the required package to use for
        runtime dependency checks) and stores them as .impname attrs on the
        Requirement instances.

        Supports ``#py..`` tags at the beginning of lines, specifying a python
        version the requirement applies to.
        """
        for line in text.split('\n'):
            line = line.strip()
            if not line:
                continue
            match = re.match(r'^#py([0-9]+) (.*)$', line)
            if match: #==> only required in given python version
                pyver, line = match.groups()
                #TODO:
                # if len(pyver) > 2:
                if not ('%s%s' % sys.version_info[:2]).startswith(pyver):
                    continue
            try:
                req, impname = line.split('#import')
            except ValueError:
                req = next(parse_requirements(line), None)
                impname = req and req.unsafe_name
            else:
                req = next(parse_requirements(req), None)
            if not req: # maybe a comment line
                continue
            req.impname = impname.strip()
            yield req

    def __init__(self, reqs, zfg=None):
        """
        Generate a list of :class:`pkg_resources.Requirement` instances.

        And additionally store all requirement specs as text with newline
        separators in the ``str`` base

        :param reqs:
            Either a single string of requirement specs or a sequence of
            strings and/or :class:`pkg_resources.Requirement` instances
        :param zfg:
            Optional zetup config object the requirements are related to
        """
        if isinstance(reqs, Requirements):
            txt = reqs.txt
            reqlist = list(self._parse(reqs.txt))
        elif isinstance(reqs, (str, unicode)):
            txt = reqs
            reqlist = list(self._parse(reqs))
        else:
            txt = ''
            reqlist = []
            for req in reqs:
                if isinstance(req, (str, unicode)):
                    reqlist.extend(self._parse(req))
                elif isinstance(req, Requirement):
                    reqlist.append(req)
                else:
                    raise TypeError(type(req))
                txt += '\n%s' % req

        self.txt = txt
        self._list = reqlist
        self.zfg = zfg

    def __eq__(self, other):
        return isinstance(other, Requirements) and (
            self._list == other._list)

    def __str__(self):
        return '\n'.join(map(str, self))

    def check(self, requirer=None, raise_=True):
        """
        Check that all requirements are satisfied.
        
        They must be importable and their versions (from their top-level
        package modules' ``.__version__`` attributes) must match the required
        specs

        Falls back to ``pkg_resources.get_distribution().version`` if
        ``.__version__`` is missing or ``None``

        :param requirer:
            Optional identification string of the requirer
        :param raise_:
            If set to ``False`` then also just return ``False`` on unsatisfied
            requirements instead of raising exceptions

        :raises zetup.DistributionNotFound:
            If a requirement can't be imported
        :raises zetup.VersionConflict:
            If a required version doesn't match
        """
        if requirer is None:
            requirer = self.zfg and '%s-%s' % (
                self.zfg.NAME, self.zfg.VERSION or '(none)')
        for req in self:
            try:
                mod = __import__(req.impname)
            except ImportError as e:
                if raise_:
                    raise DistributionNotFound(
                        req, requirer,
                        reason="%s: %s" % (type(e).__name__, e))

                return False

            if not req.specs:  # No version constraints
                continue

            try:
                version = mod.__version__
                if version is None or version == 'unknown':
                    # treat the same as if .__version__ not exists
                    # (handle in following except block)
                    # ==> same exception, just different msg
                    raise AttributeError(
                        "module's '__version__' attribute is {!r}"
                        .format(version))

            except AttributeError as e_no__version__attr:
                try:  # try to get version from distribution
                    dist = get_distribution(req.key)
                except pkg_resources.DistributionNotFound as e:
                    if raise_:
                        raise VersionConflict(
                            req, None, requirer,
                            reason="%s: %s. %s: %s" % (
                                e_no__version__attr, mod, type(e).__name__, e))

                    return False

                version = dist.version
            if version not in req:
                if raise_:
                    raise VersionConflict(req, version, requirer)

                return False

        return True

    @property
    def checked(self):
        self.check()
        return self

    def __iter__(self):
        return iter(self._list)

    def __getitem__(self, name):
        """Get a requirement by its distribution name."""
        for req in self._list:
            if name in [req.key, req.unsafe_name]:
                return req

        raise KeyError(name)

    def __delitem__(self, name):
        """Delete a requirement by its distribution name."""
        for req in list(self._list):
            if name in [req.key, req.unsafe_name]:
                return self._list.remove(req)

        raise KeyError(name)

    def __add__(self, reqs):
        """
        Create a new :class:`Requirements` instance.

        :param reqs:
            Additional requirements as string or another :class:`Requirements`
            instance
        """
        if isinstance(reqs, Requirements):
            reqs = reqs.txt
        return type(self)('%s\n%s' % (
            # For simplicity:
            # Just create explicit #import hints for every requirement:
            # '\n'.join('%s #import %s' % (req, req.impname) for req in self),
            self.txt, reqs), zfg=self.zfg)

    @property
    def py(self):
        return '%s("""\n%s\n""", zfg=zfg)' % (
            type(self).__name__, self.txt)  # '\n'.join(
            # '%s #import %s' % (req, req.impname) for req in self))

    def __repr__(self):
        return str(self)


def requires(reqs):
    """
    Create a function decorator for defining extra requirements.

    :param reqs:
        Requirement specs compatible with :class:`zetup.Requirements` 

    >>> @requires("not-installed")
    ... def func():
    ...     pass

    The actual requirements check happens only on calling the function:

    >>> func()  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    DistributionNotFound: ...func() needs not-installed ...
    <BLANKLINE>
    You can install the needed requirement with the following shell command:
    <BLANKLINE>
        pip install "not-installed"
    """
    reqs = Requirements(reqs)

    def deco(func):

        def caller(*args, **kwargs):
            reqs.check(requirer='{}.{}()'.format(
                func.__module__, func.__name__))
            return func(*args, **kwargs)

        return caller

    return deco



# ZETUP
#
# Zimmermann's Extensible Tools for Unified Project setups
#
# Copyright (C) 2014-2016 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# ZETUP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ZETUP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ZETUP. If not, see <http://www.gnu.org/licenses/>.

"""zetup.__main__

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""
from __future__ import absolute_import

import zetup.script

if __name__ == '__main__':
    zetup.script.run()



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2016 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

"""
Test :class:`zetup.class_package`.

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""

from __future__ import absolute_import

import pytest

import zetup


def test_deprecated_name():
    with pytest.warns(DeprecationWarning):
        assert zetup.classpackage is zetup.class_package


def test_bases():
    for base in [zetup.object, zetup.package]:
        assert issubclass(zetup.class_package, base)
    for nobase in [zetup.toplevel]:
        assert not issubclass(zetup.class_package, nobase)


def test_metabases():
    assert type(zetup.class_package) is zetup.meta


def test__repr__():
    assert repr(zetup.class_package) == "<class 'zetup.class_package'>"


def test_members():
    # should not contain more members than basic zetup.package
    assert dir(zetup.class_package) == dir(zetup.package)



# zetup.py
#
# Zimmermann's Python package setup.
#
# Copyright (C) 2014-2015 Stefan Zimmermann <zimmermann.code@gmail.com>
#
# zetup.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# zetup.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with zetup.py. If not, see <http://www.gnu.org/licenses/>.

import os

from zetup.process import call
from zetup.zetup import Zetup


@Zetup.command(depends=['VERSION', 'setup.py', 'zetup_config', 'tox.ini'])
def tox(self, args=None):
    return call(['tox'], env=os.environ)



# ZETUP | Zimmermann's Extensible Tools for Unified Projects
#
# Copyright (C) 2014-2019 Stefan Zimmermann <user@zimmermann.co>
#
# ZETUP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ZETUP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ZETUP. If not, see <http://www.gnu.org/licenses/>.

"""Define :class:`zetup.class_package`."""

from __future__ import absolute_import

from importlib import import_module

import zetup

__all__ = ('class_package', )


class class_package(zetup.package):
    """
    Module wrapper for a sub-package that defines a huge class.

    Enables auto-import of the class object instead of the package module
    object when imported from its parent package, and supports the
    distribution of class member definitions in separate sub-modules, given
    the following conditions:

    -   The class is named after the sub-package
    -   The class is a sub-class of :class:`zetup.object`
    -   The name of the sub-package containing the class is defined as API
        member of a parent :class:`zetup.package` instance
    -   All sub-module names are listed in the ``member_modules=`` argument,
        and members in those sub-modules are decorated using
        :meth:`zetup.meta.member` and `:meth:`zetup.meta.metamember`, which
        can be obtained from the :class:`zetup.class_member_module` wrapper

    The whole mechanism is shown below using the example ``Class`` contained
    in the ``zetup.classes.test`` package:

    ``zetup/classes/test/__init__.py`` ::

        import zetup
        zetup.package(__name__, ['Class', ...])

    ``zetup/classes/test/Class/__init__.py`` ::

        import zetup
        zetup.class_package(__name__, member_modules=['methods', ...])

        class Class:
            def __init__(self, ...):
                ...
            ...

    ``zetup/classes/test/Class/methods.py`` ::

        import zetup
        member, metamember = zetup.class_member_module(__name__ ['method'])

        @member
        def method(self, ...):
            ...

    >>> from zetup.classes.test import Class
    >>> Class
    <class 'zetup.classes.test.Class'>
    >>> Class.method
    <... Class.method...>
    """

    __package__ = zetup

    def __init__(self, pkgname, member_modules=None):
        """Created with ``__name__`` of the subpackage defining the class
        and the optional list of `membermodules` to automatically import
        (sub-modules defining additional class members).
        """
        parentpkgname, classname = pkgname.rsplit('.', 1)
        super(class_package, self).__init__(pkgname, [classname])

        pkgcls = type(self)
        pkgdict = self.__dict__

        def load_class():
            """Get the actual class object from this class package
            and automatically import all defined ``membermodules``.
            """
            classobj = pkgcls.__getattribute__(self, classname)
            if not getattr(classobj, '__package__', None):
                classobj.__package__ = parentpkgname
            if member_modules is not None:
                for modname in member_modules:
                    mod = import_module('%s.%s' % (pkgname, modname))
                    mod.owner = classobj
            return classobj

        class package(type(self)):
            """Help tools like sphinx ``.. autoclass::`` doc generator
            by delegating (almost) all attributes to the actual class object.

            * ``.. autoclass:: package.Class`` first looks up
              ``package.Class`` in imported modules and therefore
              doesn't get the actual class object itself.
            """
            def __getattribute__(self, name):
                if name == classname:
                    # get the actual class object
                    try:
                        return pkgdict[classname]
                    except KeyError:
                        # only load once
                        classobj = pkgdict[classname] = load_class()
                        return classobj
                # only take these few special attributes
                # from the classpackage instance itself
                if name in [
                        '__name__', '__all__',  '__module__', '__path__',
                        '__file__', '__class__', '__dict__',
                ]:
                    return pkgcls.__getattribute__(self, name)
                # and take all other attributes from the class object
                # (by first recursively calling this __getattribute__)
                return getattr(getattr(self, classname), name)

            def __repr__(self):
                return "<%s for %s from %s>" % (
                    pkgcls.__name__, repr(getattr(self, classname)),
                    repr(self.__file__))

        self.__class__ = package



# ZETUP | Zimmermann's Extensible Tools for Unified Projects
#
# Copyright (C) 2014-2019 Stefan Zimmermann <user@zimmermann.co>
#
# ZETUP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ZETUP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ZETUP. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import sys
from inspect import ismodule
try:
    from inspect import signature
except ImportError:
    def signature(func):
        return ""

import zetup
from .object import object

__all__ = ('apifunction', )


class apifunction(object):

    __package__ = zetup.__package__

    def __init__(self, pkg, func=None):
        if not ismodule(pkg):
            raise TypeError(
                "{!r} must be initialized with a module object as first arg"
                ", not {!r}".format(type(self), pkg))

        if func is None:
            self.__package__ = pkg
        else:
            self.__package__ = pkg.__name__
            self.__module__ = func.__module__
            self.__class__ = type('apifunction', (type(self), ), {
                '__doc__': "", '__call__': staticmethod(func)})
            self.__func__ = func

            mod = sys.modules[self.__module__]
            if not hasattr(mod, '__test__'):
                mod.__test__ = {}
            mod.__test__[func.__name__] = func.__doc__

    def __call__(self, func):
        return type(self)(self.__package__, func)

    def __getattr__(self, name):
        if '__func__' in self.__dict__:
            try:
                return getattr(self.__func__, name)

            except AttributeError:
                pass
        raise AttributeError(
            "{!r} has no attribute {!r}".format(self, name))

    def __str__(self):
        if '__func__' in self.__dict__:
            return '.'.join((self.__package__, getattr(
                self.__func__, '__qualname__', self.__func__.__name__)))

        return repr(self)

    def __repr__(self):
        if '__func__' in self.__dict__:
            return "<{} {}.{}{}>".format(
                type(self).__name__, self.__package__, getattr(
                    self.__func__, '__qualname__', self.__func__.__name__),
                signature(self.__func__))

        return "<{} decorator for {!r}>".format(
            type(self), self.__package__)



"""Test :mod:`zetup.pip.errors`."""

import zetup
from zetup.pip.errors import ZetupPipError


class TestZetupPipError:
    """Test :exc:`zetup.pip.errors.ZetupPipError`."""

    def test_toplevel_export(self):
        """Should be properly provided by :mod:`zetup` API."""
        assert zetup.ZetupPipError is ZetupPipError

    def test_superclass(self):
        """Should be derived from :exc:`zetup.ZetupError`."""
        assert issubclass(ZetupPipError, zetup.ZetupError)

    def test_attributes(self):
        """Should store ``.command`` and ``.exit_code``."""
        exc = ZetupPipError(['invalid'], 1)
        assert exc.command == ['invalid']
        assert exc.exit_code is 1

    def test_message(self):
        """Should provide correct exception message."""
        exc = ZetupPipError(['invalid'], 1)
        assert str(exc) == "pip command ['invalid'] failed with exit code 1"



"""Define :func:`zetup.pip` command runner and :exc:`zetup.ZetupPipError`."""

import logging

from zetup.requires import requires

from .errors import ZetupPipError

__all__ = ('ZetupPipError', 'pip')


@requires("pip >= 10.0")
def pip(command, raise_=True):
    """
    Run a pip `command`.

    :param command: The list of pip arguments

    >>> import zetup

    >>> zetup.pip(['list', '-v'])
    Package...Version...Location...Installer
    -------...-------...--------...---------
    ...
    0

    :return:
        ``0`` on success or, if called with ``raise_=False``, also any
        non-zero pip exit code

    >>> zetup.pip(['invalid'], raise_=False)
    1

    :raises zetup.ZetupPipError:
        on non-zero pip exit codes, if not called with ``raise_=False``

    >>> zetup.pip(['invalid'])
    Traceback (most recent call last):
    ...
    ZetupPipError: pip command ['invalid'] failed with exit code 1
    """
    from pip._internal import configuration, main

    # HACK: Avoid unnecessary debug messages (which also destroy doctests)
    configuration.logger.level = logging.INFO
    try:
        exit_code = main(command)
    except SystemExit as exc:
        exit_code = exc.code
    if raise_ and exit_code != 0:
        raise ZetupPipError(command, exit_code)

    return exit_code



"""High-level ZFG file handling."""

from .data import ZFGData
from .parser import parse_zfg

__all__ = ('ZFGFile', )


class ZFGFile(ZFGData):
    """
    The actual ZFG file handler.

    >>> from zetup import ZFGFile

    >>> import os
    >>> zfg = ZFGFile(os.path.join(os.path.dirname(__file__), 'test.zfg'))
    >>> zfg
    ZFGFile '...test.zfg'
    >>> zfg.path
    '...test.zfg'

    Parse the file text on :meth:`.load`, which can then be accessed and
    changed via the basic :class:`zetup.ZFGData` API:

    >>> zfg.load()
    >>> zfg.sections()
    ['test section', 'other test section']

    Save back changed data with :meth:`.save`, which preserves the original
    user-style of the input text as much as possible
    """

    def __init__(self, path):
        """Initialize with a ZFG file `path`."""
        super(ZFGFile, self).__init__()
        self.path = path

    def load(self):
        """Parse the ZFG file text."""
        with open(self.path) as file:
            parse_zfg(file.read(), zfg=self)

    def save(self):
        """Save data back to the ZFG file."""
        with open(self.path, 'w') as file:
            file.write(str(self))

    def __repr__(self):
        """Show ZFG file path and loaded content."""
        text = "{} {!r}".format(type(self).__name__, self.path)
        if self.lines:
            text += "\n{}".format(super(ZFGFile, self).__repr__())
        return text



# ZETUP | Zimmermann's Extensible Tools for Unified Projects
#
# Copyright (C) 2014-2019 Stefan Zimmermann <user@zimmermann.co>
#
# ZETUP is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ZETUP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ZETUP. If not, see <http://www.gnu.org/licenses/>.

"""
Several ``module`` object wrappers.

For modules, packages, top-level packages, and top-level packages for extra
features

.. moduleauthor:: Stefan Zimmermann <zimmermann.code@gmail.com>
"""

from __future__ import absolute_import

import sys
from collections import MutableMapping
from importlib import import_module
from inspect import ismodule
from itertools import chain
from types import ModuleType
from warnings import warn

import zetup
from .object import object, meta
from .annotate import annotate, annotate_extra
from .doc import AutoDocScopeModule

__all__ = ('extra_toplevel', 'module', 'package', 'toplevel')


class module__dict__(dict):

    __base__ = None

    def items(self):
        if self.__base__ is None:
            return dict.items(self)

        return chain(dict.items(self), self.__base__.items())


MODULE__DICT__CACHE = {}


class deprecated(str):
    """
    Simple ``str`` class wrapper for indicating deprecation.

    Marks a :class:`zetup.module` API member (alias) name as deprecated
    """

    def __repr__(self):
        return "deprecated(%s)" % str.__repr__(self)


class module(ModuleType, object):
    """
    Basic module object wrapper.

    Providing clean dynamic API imports from modules
    """

    __package__ = zetup

    @property
    def __dict__(self):
        return MODULE__DICT__CACHE.setdefault(self, module__dict__())

    def __init__(
            self, name, __all__=None,
            aliases=None, deprecated_aliases=None,
            __getitem__=None, __iter__=None, __call__=None):
        """
        Wrap a package module given by `name`.

        Original package module object is replaced in ``sys.modules`` and
        stored in :attr:``.__module__``

        Optional `__all__` list defines the package API

        Optional `aliases` and `deprecated_aliases` map alternative names to
        API names

        `__getitem__`, `__iter__`, and `__call__` features can be added to the
        package wrapper by providing handler functions or other callable
        objects (which are not called with a ``self`` argument)
        """
        mod = sys.modules[name]
        ModuleType.__init__(  # pylint: disable=no-member
            self, name, mod.__doc__)

        __dict__ = MODULE__DICT__CACHE.setdefault(self, module__dict__())
        __dict__.__base__ = mod.__dict__

        self.__name__ = name
        self.__module__ = mod
        sys.modules[name] = self

        self.__dict__['__all__'] = api \
            = dict.fromkeys(__all__) if __all__ is not None else {}
        if aliases is not None:
            api.update(aliases)
        if deprecated_aliases is not None:
            api.update(
                (deprecated(alias), name)
                for alias, name in dict(deprecated_aliases).items())

        # if api is not None:
        #     for submodname, members in dict(__all__).items():
        #         self.__dict__['__all__'].update(
        #             (name, submodname) for name in members)

        cls = type(self)
        cls.__getitem__.funcs[self] = __getitem__
        cls.__iter__.funcs[self] = __iter__
        cls.__call__.funcs[self] = __call__

    def __getitem__(self, key):
        cls = type(self)
        func = cls.__getitem__.funcs.get(self)
        if func is None:
            raise TypeError(
                "{!r} is not subscriptable. "
                "Instantiate {!r} with __getitem__=<func> to change that."
                .format(self, cls))

        return func(key)

    __getitem__.funcs = {}

    def __iter__(self):
        cls = type(self)
        func = cls.__iter__.funcs.get(self)
        if func is None:
            raise TypeError(
                "{!r} is not iterable. "
                "Instantiate {!r} with __iter__=<func> to change that."
                .format(self, cls))

        return func()

    __iter__.funcs = {}

    def __call__(self, *args, **kwargs):
        cls = type(self)
        func = cls.__call__.funcs.get(self)
        if func is None:
            raise TypeError(
                "{!r} is not callable. "
                "Instantiate {!r} with __call__=<func> to change that."
                .format(self, cls))

        return func(*args, **kwargs)

    __call__.funcs = {}

    @property
    def __all__(self):
        """Get API names list (without deprecated aliases)."""
        return list(set(chain(
            getattr(self.__module__, '__all__', ()), (
                name for name in self.__dict__['__all__']
                if not isinstance(name, deprecated)))))

    def __setattr__(self, name, value):
        """
        Prevent submodules from being added as attributes.

        To avoid unnecessary ``dir()`` pollution
        """
        if isinstance(value, AutoDocScopeModule) or ismodule(value) and (
                value.__name__ == '%s.%s' % (self.__name__, name) and
                not isinstance(value, package)):
            return

        if isinstance(value, zetup.class_package):
            value = getattr(value, name)
        object.__setattr__(self, name, value)

    def __getattribute__(self, name):
        """Dynamically access API from wrapped module or import extra API."""
        if name.startswith('__'):
            try:
                return object.__getattribute__(self, name)

            except AttributeError:
                pass

        # get actual API key object corresponding to given name...
        for alias, realname in self.__dict__['__all__'].items():
            if name == alias:
                name = alias
                break
        else:
            realname = None

        # ... to check if name is defined as alias
        if realname is not None:
            if isinstance(name, deprecated):
                warn("%s.%s is deprecated in favor of %s.%s" % (
                    self.__name__, name, self.__name__, realname
                ), DeprecationWarning)
            name = realname

        try:  # first try to get attr from wrapped original module
            return getattr(self.__module__, name)

        except AttributeError:
            try:  # then from wrapper module
                if name in type(self).__dict__:
                    return object.__getattribute__(self, name)

                return self.__dict__[name]

            except KeyError:
                if name in getattr(self.__module__, '__all__', ()):
                    raise AttributeError(
                        "{!r} has no attribute {!r} although listed in "
                        "__all__".format(self.__module__, name))

        raise AttributeError(
            "{!r} has no attribute {!r}".format(self, name))

    def __dir__(self):
        """Additionally get all API member names."""
        def exclude():
            """
            Get names of submodules implicitly added to ``__dict__``.

            Which happens on ``from .submodule import ...`` statements in
            wrapped module following the wrapper instantiation
            """
            for name, obj in self.__dict__.items():
                if name.startswith('__'):
                    continue

                if ismodule(obj) and not isinstance(obj, package):
                    yield name

        return list(chain(
            set(
                object.__dir__(self)  # pylint: disable=no-member
            ).difference(exclude()),
            self.__all__))

    def __repr__(self):
        """Create module-style representation."""
        return "<%s %s from %s>" % (
            type(self).__name__, repr(self.__name__),
            repr(self.__module__.__file__))


class package(module):
    """
    Package module object wrapper.

    Providing clean dynamic API imports from sub-modules
    """

    __package__ = zetup

    def __getattribute__(self, name):
        try:
            obj = super(package, self).__getattribute__(name)

        except AttributeError as exc:
            try:  # try to find a matching submodule
                obj = import_module('%s.%s' % (self.__name__, name))
            except ImportError:
                raise AttributeError(
                    "%s has no attribute %s" % (repr(self), repr(name)))

        if isinstance(obj, zetup.class_package):
            classobj = getattr(obj, name)
            setattr(self, name, classobj)
            return classobj

        return obj


class toplevel(package):
    """
    Special top-level package module object wrapper.

    For clean dynamic API import from sub-modules and automatic application
    of func:`zetup.annotate`
    """

    __package__ = zetup

    def __init__(
            self, name, __all__=None,
            aliases=None, deprecated_aliases=None,
            check_requirements=True, check_packages=True,
            __getitem__=None, __iter__=None, __call__=None):
        """
        Wrap top-level package module given by `name` and `api` list.

        See :class:`zetup.package` for details about defining the package
        API and special features

        See :func:`zetup.annotate` for details about the check options
        """
        super(toplevel, self).__init__(
            name, __all__,
            aliases=aliases, deprecated_aliases=deprecated_aliases,
            __getitem__=__getitem__, __iter__=__iter__, __call__=__call__)

        zfg = annotate(
            name, check_requirements=check_requirements,
            check_packages=check_packages)
        self.__package__ = pkg = zfg.PACKAGES[name]
        pkg.zetup_config = zfg


class extra_toplevel_meta(meta):
    """Metaclass for :class:`zetup.extra_toplevel`."""

    extra = None

    def __getitem__(cls, extra):
        mcs = type(cls)
        meta = type('%s[%s]' % (mcs.__name__, extra), (mcs, ), {
            'extra': extra,
        })
        return meta('%s[%s]' % (cls.__name__, extra), (cls, ), {})


class extra_toplevel(
        # PY2/3 compatible way to assign metaclass
        extra_toplevel_meta('extra_toplevel_base', (package, ), {})):
    """
    Special extra feature top-level package module object wrapper.

    For clean dynamic API import from sub-modules and automatic application
    of func:`zetup.annotate_extra`
    """

    __package__ = zetup

    def __init__(
            self, toplevel, name, __all__=None,
            aliases=None, deprecated_aliases=None,
            check_requirements=True,
            __getitem__=None, __iter__=None, __call__=None):
        """
        Wrap top-level package module given by `name` and `api` list.

        See :class:`zetup.package` for details about defining the package
        API and special features

        See :func:`zetup.annotate_extra` for details about the check option
        """
        super(extra_toplevel, self).__init__(
            name, __all__,
            aliases=aliases, deprecated_aliases=deprecated_aliases,
            __getitem__=__getitem__, __iter__=__iter__, __call__=__call__)

        extra = type(self).extra
        annotate_extra[extra](  # pylint: disable=unsubscriptable-object
            toplevel, name, check_requirements=check_requirements)



"""ZFG data structures."""

import json
from collections import OrderedDict

from zetup.object import object

from .line import OptionLine, SectionLine

__all__ = ('ZFGData', )


class ZFGEncoder(json.JSONEncoder):

    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, ZFGSection):
            return OrderedDict(obj)

        super(ZFGEncoder, self).default(obj)


class ZFGData(object):
    """The basic manager class for parsed ZFG file data."""

    #: The parsed top-level lines of a ``.zfg`` file.
    lines = None

    def __init__(self):
        self.lines = []

    def __getitem__(self, key):
        if key not in self.sections():
            raise KeyError(key)

        return ZFGSection(key, data=self)

    def sections(self):
        """Get a list of all root-level sections."""
        result = []
        for parsed_line in self.lines:
            if isinstance(parsed_line, SectionLine):
                name = parsed_line.name.split(':')[0]
                if name not in result:
                    result.append(name)
        return result

    def to_zfg(self):
        return "\n".join(line.to_zfg() for line in self.lines)

    def __str__(self):
        return self.to_zfg()

    def __repr__(self):
        return "{}\n{}".format(type(self).__name__, json.dumps(
            OrderedDict((
                (name, OrderedDict(self[name]))
                for name in self.sections())),
            cls=ZFGEncoder, indent=4))


class ZFGSection(object):

    def __init__(self, name, data):
        self.name = name
        self.data = data

    def subsections(self):
        result = []
        for line in self.data.lines:
            if isinstance(line, SectionLine) and ':' in line.name:
                name, subname = line.name.split(':', 1)
                if name == self.name:
                    result.append(subname.split(':')[0])
        return result

    def options(self):
        result = []
        for line in self.data.lines:
            if isinstance(line, SectionLine):
                if line.name != self.name:
                    continue

                for line in line.lines:
                    if isinstance(line, OptionLine):
                        result.append(line.name)
        return result

    def __iter__(self):
        for line in self.data.lines:
            if isinstance(line, SectionLine):
                if line.name == self.name:
                    for line in line.lines:
                        if isinstance(line, OptionLine):
                            yield (line.name, line.value)

                elif ':' in line.name:
                    name, subname = line.name.rsplit(':', 1)
                    if name == self.name:
                        yield (subname, type(self)(line.name, self.data))

    def __getitem__(self, key):
        for name, data in self:
            if name == key:
                return data

        raise KeyError(key)

    def __repr__(self):
        return json.dumps(OrderedDict(self), indent=4)



"""Define exceptions for failing ``zetup.pip`` calls."""

from zetup.error import ZetupError


class ZetupPipError(ZetupError):
    """
    ``zetup.pip([command...])`` failed.
    
    ``.command`` argument list and ``.exit_code`` can be retrieved from the
    exception object:

    >>> import zetup

    >>> try:
    ...     zetup.pip(['invalid'])
    ... except zetup.ZetupPipError as exc:
    ...     exc.command, exc.exit_code
    (['invalid'], 1)
    """

    def __init__(self, command, exit_code):
        """
        Compose error message from `command` list and `exit_code`.

        Also store the parameters as exception object attributes
        """
        ZetupError.__init__(
            self, "pip command {!r} failed with exit code {:d}".format(
                command, exit_code))
        self.command = command
        self.exit_code = exit_code
